/**
  @file utils.c

  @author Johan Hedberg <johan.hedberg@nokia.com>

  Copyright (C) 2004-2006 Nokia Corporation. All rights reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License, version 2.1, as published by the Free Software Foundation.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

*/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <glib.h>
#include <termios.h>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "log.h"
#include "utils.h"

#ifdef DEBUG
char *bytestr(const uint8_t *uuid, int len) {
    int i;
    char *str = g_malloc((len << 1) + 1);

    for (i = 0; i < len; i++)
        sprintf(str + (2*i), "%02X", uuid[i]);

    return str;
}
#endif

gboolean fd_raw_mode(int fd) {
    struct termios mode;

    memset(&mode, 0, sizeof (mode));
    if (tcgetattr(fd, &mode) < 0) {
        debug("tcgetattr(%d, &mode): %s", fd, strerror(errno));
        return FALSE;
    }

    mode.c_iflag = 0;
    mode.c_oflag &= ~OPOST;
    mode.c_lflag &= ~(ISIG | ICANON | ECHO
#ifdef XCASE
            | XCASE
#endif
            );
    mode.c_cc[VMIN] = 1;
    mode.c_cc[VTIME] = 0;

    if (tcsetattr(fd, TCSADRAIN, &mode) < 0) {
        debug("tcsetattr(%d, TCSADRAIN, &mode): %s", fd, strerror(errno));
        return FALSE;
    }

    return TRUE;
}

glong get_uname(gunichar2 **uname, const gchar *name) {
    glong uname_len;

    if (*name == '\0') {
        *uname = NULL;
        return 0;
    }

    *uname = g_utf8_to_utf16(name, -1, NULL, &uname_len, NULL);

    if (*uname == NULL)
        uname_len = -1;
    else {
        int i;
        /* g_utf8_to_utf16 produces host-byteorder UTF-16,
         * but OBEX requires network byteorder (big endian) */
        for (i = 0; i < uname_len; i++)
            (*uname)[i] = g_htons((*uname)[i]);
        uname_len = (uname_len + 1) << 1;
    }

    return uname_len;
}

char * make_iso8601(time_t time) {
    GTimeVal tv;

    tv.tv_sec = time;
    tv.tv_usec = 0;

    return g_time_val_to_iso8601(&tv);
}

time_t parse_iso8601(const gchar *str, int len) {
    gchar    *tstr;
    GTimeVal  tv;

    /* According to spec the time doesn't have to be null terminated */
    if (str[len - 1] != '\0')
        tstr = g_strndup(str, len);
    else
        tstr = g_strdup(str);

    if (g_time_val_from_iso8601(tstr, &tv) == FALSE) {
        g_free (tstr);
        return -1;
    }

    g_free (tstr);
    return tv.tv_sec;
}

